@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2015 David Robillard <http://drobilla.net>
+  Copyright 2011-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -87,6 +87,7 @@ struct SerdReaderImpl {
 	bool              paging;     ///< True iff reading a page at a time
 	bool              strict;     ///< True iff strict parsing
 	bool              eof;
+	bool              error;
 	bool              seen_genid;
 #ifdef SERD_STACK_CHECK
 	Ref*              allocs;     ///< Stack of push offsets
@@ -115,7 +116,14 @@ page(SerdReader* reader)
 	if (n_read == 0) {
 		reader->file_buf[0] = '\0';
 		reader->eof         = true;
-		return ferror(reader->fd) ? SERD_ERR_UNKNOWN : SERD_FAILURE;
+		if (ferror(reader->fd)) {
+			reader->error = true;
+			r_err(reader, SERD_ERR_UNKNOWN,
+			      "read error: %s\n", strerror(errno));
+			return SERD_ERR_UNKNOWN;
+		} else {
+			return SERD_FAILURE;
+		}
 	} else if (n_read < SERD_PAGE_SIZE) {
 		reader->file_buf[n_read] = '\0';
 	}
@@ -173,8 +181,8 @@ static Ref
 push_node_padded(SerdReader* reader, size_t maxlen,
                  SerdType type, const char* str, size_t n_bytes)
 {
-	uint8_t* mem = serd_stack_push(&reader->stack,
-	                               sizeof(SerdNode) + maxlen + 1);
+	void* mem = serd_stack_push_aligned(
+		&reader->stack, sizeof(SerdNode) + maxlen + 1, sizeof(SerdNode));
 
 	SerdNode* const node = (SerdNode*)mem;
 	node->n_bytes = node->n_chars = n_bytes;
@@ -182,13 +190,13 @@ push_node_padded(SerdReader* reader, size_t maxlen,
 	node->type    = type;
 	node->buf     = NULL;
 
-	uint8_t* buf = mem + sizeof(SerdNode);
+	uint8_t* buf = (uint8_t*)(node + 1);
 	memcpy(buf, str, n_bytes + 1);
 
 #ifdef SERD_STACK_CHECK
 	reader->allocs = realloc(
 		reader->allocs, sizeof(uint8_t*) * (++reader->n_allocs));
-	reader->allocs[reader->n_allocs - 1] = (mem - reader->stack.buf);
+	reader->allocs[reader->n_allocs - 1] = ((uint8_t*)mem - reader->stack.buf);
 #endif
 	return (uint8_t*)node - reader->stack.buf;
 }
@@ -243,7 +251,7 @@ pop_node(SerdReader* reader, Ref ref)
 #endif
 		SerdNode* const node = deref(reader, ref);
 		uint8_t* const  top  = reader->stack.buf + reader->stack.size;
-		serd_stack_pop(&reader->stack, top - (uint8_t*)node);
+		serd_stack_pop_aligned(&reader->stack, top - (uint8_t*)node);
 	}
 	return 0;
 }
@@ -761,7 +769,7 @@ read_LANGTAG(SerdReader* reader)
 	}
 	Ref ref = push_node(reader, SERD_LITERAL, "", 0);
 	push_byte(reader, ref, eat_byte_safe(reader, c));
-	while ((c = peek_byte(reader)) && in_range(c, 'a', 'z')) {
+	while ((c = peek_byte(reader)) && is_alpha(c)) {
 		push_byte(reader, ref, eat_byte_safe(reader, c));
 	}
 	while (peek_byte(reader) == '-') {
@@ -1237,6 +1245,7 @@ read_predicateObjectList(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 			return true;
 		}
 
+		bool ate_semi = false;
 		do {
 			read_ws_star(reader);
 			switch (c = peek_byte(reader)) {
@@ -1246,8 +1255,13 @@ read_predicateObjectList(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 				return true;
 			case ';':
 				eat_byte_safe(reader, c);
+				ate_semi = true;
 			}
 		} while (c == ';');
+
+		if (!ate_semi) {
+			return r_err(reader, SERD_ERR_BAD_SYNTAX, "missing ';' or '.'\n");
+		}
 	}
 
 	pop_node(reader, ctx.predicate);
@@ -1446,7 +1460,7 @@ read_statement(SerdReader* reader)
 	switch (peek_byte(reader)) {
 	case '\0':
 		reader->eof = true;
-		return true;
+		return !reader->error;
 	case '@':
 		TRY_RET(read_directive(reader));
 		read_ws_star(reader);
@@ -1472,7 +1486,7 @@ read_turtleDoc(SerdReader* reader)
 	while (!reader->eof) {
 		TRY_RET(read_statement(reader));
 	}
-	return true;
+	return !reader->error;
 }
 
 SERD_API
@@ -1486,33 +1500,18 @@ serd_reader_new(SerdSyntax        syntax,
                 SerdEndSink       end_sink)
 {
 	const Cursor cur = { NULL, 0, 0 };
-	SerdReader*  me  = (SerdReader*)malloc(sizeof(struct SerdReaderImpl));
+	SerdReader*  me  = (SerdReader*)calloc(1, sizeof(SerdReader));
 	me->handle           = handle;
 	me->free_handle      = free_handle;
 	me->base_sink        = base_sink;
 	me->prefix_sink      = prefix_sink;
 	me->statement_sink   = statement_sink;
 	me->end_sink         = end_sink;
-	me->error_sink       = NULL;
-	me->error_handle     = NULL;
 	me->default_graph    = SERD_NODE_NULL;
-	me->fd               = 0;
 	me->stack            = serd_stack_new(SERD_PAGE_SIZE);
 	me->syntax           = syntax;
 	me->cur              = cur;
-	me->bprefix          = NULL;
-	me->bprefix_len      = 0;
 	me->next_id          = 1;
-	me->read_buf         = 0;
-	me->file_buf         = 0;
-	me->read_head        = 0;
-	me->strict           = false;
-	me->eof              = false;
-	me->seen_genid       = false;
-#ifdef SERD_STACK_CHECK
-	me->allocs           = 0;
-	me->n_allocs         = 0;
-#endif
 
 	me->rdf_first = push_node(me, SERD_URI, NS_RDF "first", 48);
 	me->rdf_rest  = push_node(me, SERD_URI, NS_RDF "rest", 47);
